/*******************************************************************
 *
 *	File:		CFileReader.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 2002, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#define COFF_IMAGE_WITH_PE
#include "coff_pe/arm.h"
#include "coff_pe/pe.h"

#undef _LIT

#include <e32def.h>
#include <e32uid.h>
#include <f32file.h>
#include "debug.h"

#define COMPRESSED_IMAGE_FILE
#include "CFileReader.h"
#include "ECompXLFileFormat.h"
#include "Panic.h"
#include <EZDecompressor.h>
#include <EZGzip.h>


#ifdef COMPRESSED_IMAGE_FILE


void TGzipReader::ReadFile(RFile& aFile)
{
	DPRINT((HERE("TGzipReader::ReadFile")));

	PanicIfError(aFile.Read(iReadBuf), ECompXLPanicCouldNotReadPartOfImage);
	iZLibStream.next_in  = (TUint8*) iReadBuf.Ptr();
	iZLibStream.avail_in = iReadBuf.Size();

	DPRINT((HERE("TGzipReader::ReadFile - done")));
}

void TGzipReader::Inflate(RFile& aFile, TDes8& aBuf)
{
	DPRINT((HERE("TGzipReader::Inflate: readSize=%d"), aBuf.MaxSize()));

	iZLibStream.next_out = (TUint8*) aBuf.Ptr();
	iZLibStream.avail_out = aBuf.MaxSize();
	TInt err = 0;
	do
	{
		if (iZLibStream.avail_in == 0)
			ReadFile(aFile);
		err = inflate(&iZLibStream, Z_NO_FLUSH);
	} while(!err && (iZLibStream.avail_out!=0));

	if (iZLibStream.avail_out!=0)
		Panic(ECompXLPanicInflateError);

	DPRINT((HERE("TGzipReader::Inflate - done, err=%d"), err));
}

void TGzipReader::Construct(RFile& aFile)
{
	DPRINT((HERE("TGzipReader::Construct")));

	iZLibStream.zalloc = (alloc_func)0;
	iZLibStream.zfree = (free_func)0;
	iZLibStream.opaque = (voidpf)0;
	iZLibStream.next_in  = (TUint8*) iReadBuf.Ptr();
	iZLibStream.avail_in = iReadBuf.Size();
	iZLibStream.next_out = NULL;
	iZLibStream.avail_out = 0;

	ReadFile(aFile);

	TInt err = inflateInit2(&iZLibStream, -MAX_WBITS);
	PanicIfError(err, ECompXLPanicInflateInit2Failed);

	DPRINT((HERE("TGzipReader::Construct - done")));
}

void TGzipReader::CheckEnd()
{
	DPRINT((HERE("TGzipReader::CheckEnd - ")));

		/*
		 * Frees any resources allocated by zlib
		 */
    PanicIfError(inflateEnd(&iZLibStream), ECompXLPanicInflateEndFailed);

	DPRINT((HERE("TGzipReader::CheckEnd - done")));
}


CGZipPEFileReader* CGZipPEFileReader::NewL()
{
	CGZipPEFileReader*	reader = new(ELeave)CGZipPEFileReader;
	CleanupStack::PushL(reader);
	reader->ConstructL();
	CleanupStack::Pop(reader);
	return reader;
}

CGZipPEFileReader::~CGZipPEFileReader()
{
	DPRINT((HERE("CGZipPEFileReader::~CGZipPEFileReader")));

	iFile.Close();
	iFs.Close();

	DPRINT((HERE("CGZipPEFileReader::~CGZipPEFileReader - done")));
}


void CGZipPEFileReader::OpenImageL()
{
		/*
		 * For target builds, we open our own executable image (either the .exe or
		 * the .app dll). For WINS we use a fixed test exe/app.
		 */
#ifdef ECOMPXL_EXE
	#ifdef __WINS__
		TFileName	me = _L("c:\\system\\programs\\hello.exe");
	#else
		TFileName	me = RProcess().FileName();
	#endif
#else
	#ifdef __WINS__
		TFileName	me = _L("c:\\system\\apps\\hello\\hello.app");
	#else
		TFileName	me;
		Dll::FileName(me);
	#endif
#endif

	PanicIfError(iFile.Open(iFs, me, 0), ECompXLPanicCouldNotOpenOwnImageFile);
}


void CGZipPEFileReader::ConstructL()
{
	DPRINT((HERE("CGZipPEFileReader::ConstructL")));

	PanicIfError(iFs.Connect(), ECompXLPanicCouldNotConnectToFileServer);
	OpenImageL();

	DPRINT((HERE("CGZipPEFileReader::ConstructL - done")));
}


CFileReaderBase* CreateFileReaderL()
{
	return CGZipPEFileReader::NewL();
}


static void SetBlocks(TSectionTableEntry* aSectionTable, TUint8* aImageVma)
{
	aSectionTable->iStartBlock = (aSectionTable->iImageVma - aImageVma) / KPageSize;
	aSectionTable->iLastBlock = aSectionTable->iStartBlock + ((aSectionTable->iSize + KPageSize -1) / KPageSize);
}


void CGZipPEFileReader::LoadHeadersL(TImageInfo& aImageInfo, TSectionTableEntry* aSectionTable)
{
	DPRINT((HERE("CGZipPEFileReader::ReadHeadersL")));

	TInt		offset;
	TPtr8		p((TUint8*)&offset, sizeof(offset));

		/*
		 * The last 32-bit word in our E32 image file will tell us where in the
		 * E32 image file we've stored the gzipped PE image
		 */
	TInt		fileSize;
	PanicIfError(iFile.Size(fileSize), ECompXLPanicCouldNotQueryImageFileSize);
	TInt	endMinus4 = fileSize - 4;
	PanicIfError(iFile.Seek(ESeekStart, endMinus4), ECompXLPanicCouldNotSeekToLastDWord);
	PanicIfError(iFile.Read(p), ECompXLPanicCouldNotReadLastDWord);

		/*
		 * Seek to correct offset
		 */
	if ((offset<0) || (offset > fileSize))
		Panic(ECompXLPanicGZipFileOffsetOutOfRange);
	PanicIfError(iFile.Seek(ESeekStart, offset), ECompXLPanicCouldNotSeekToGZipFileOffset);

		/*
		 * Read gzip header
		 */
	TEZGZipHeader		gzipHeader;
	TRAPD(err, EZGZipFile::ReadHeaderL(iFile, gzipHeader));
	PanicIfError(err, ECompXLPanicCouldNotReadGzipHeader);

	iGzipReader.Construct(iFile);

	TECompXLFileHeader	eCompXLFileHeader;
	TPtr8	p1((TUint8*)&eCompXLFileHeader, sizeof(eCompXLFileHeader));
	iGzipReader.Inflate(iFile, p1);

	if (eCompXLFileHeader.iMagic != ECOMPXL_MAGIC)
		Panic(ECompXLPanicInvalidMagicInGZipXLHeader);
	if (eCompXLFileHeader.iVersion != KECompXLVersion1)
		Panic(ECompXLPanicInvalidVersionInGZipXLHeader);
		
	aImageInfo.iImageBase					= (TUint8*) eCompXLFileHeader.iImageBaseVma;
	aImageInfo.iImageEntyPoint				= (TUint8*) eCompXLFileHeader.iImageEntryPoint;

	aSectionTable[KSectionCode].iImageVma	= (TUint8*) eCompXLFileHeader.iCodeSectionVma;
	aSectionTable[KSectionCode].iSize		= eCompXLFileHeader.iCodeSectionSize;
	SetBlocks(&aSectionTable[KSectionCode], aImageInfo.iImageBase);

	aSectionTable[KSectionBss].iImageVma	= (TUint8*) eCompXLFileHeader.iBssSectionVma;
	aSectionTable[KSectionBss].iSize		= eCompXLFileHeader.iBssSectionSize;
	SetBlocks(&aSectionTable[KSectionBss], aImageInfo.iImageBase);

	aSectionTable[KSectionData].iImageVma	= (TUint8*) eCompXLFileHeader.iDataSectionVma;
	aSectionTable[KSectionData].iSize		= eCompXLFileHeader.iDataSectionSize;
	SetBlocks(&aSectionTable[KSectionData], aImageInfo.iImageBase);

	aSectionTable[KSectionImport].iImageVma = (TUint8*) eCompXLFileHeader.iImportSectionVma;
	aSectionTable[KSectionImport].iSize		= eCompXLFileHeader.iImportSectionSize;
	SetBlocks(&aSectionTable[KSectionImport], aImageInfo.iImageBase);

	aSectionTable[KSectionReloc].iImageVma	= (TUint8*) eCompXLFileHeader.iRelocSectionVma;
	aSectionTable[KSectionReloc].iSize		= eCompXLFileHeader.iRelocSectionSize;
	SetBlocks(&aSectionTable[KSectionReloc], aImageInfo.iImageBase);

	DPRINT((HERE("CGZipPEFileReader::ReadHeadersL - done")));
}

void CGZipPEFileReader::GunzipSectionL(const TSectionTableEntry& aSectionEntry)
{
	DPRINT((HERE("CGZipPEFileReader::GunzipSectionL")));

	TPtr8	p(aSectionEntry.iLoadedVma, aSectionEntry.iSize);
	iGzipReader.Inflate(iFile, p);

	DPRINT((HERE("CGZipPEFileReader::GunzipSectionL - done")));
}


void CGZipPEFileReader::LoadSectionsL(const TSectionTableEntry* aSectionTable)
{
	DPRINT((HERE("CGZipPEFileReader::LoadSectionsL")));

	if (aSectionTable[KSectionCode].iLoadedVma)
	{
		DPRINT((HERE("CGZipPEFileReader::LoadSectionsL - code section")));
		GunzipSectionL(aSectionTable[KSectionCode]);
	}

	if (aSectionTable[KSectionData].iLoadedVma)
	{
		DPRINT((HERE("CGZipPEFileReader::LoadSectionsL - data section")));
		GunzipSectionL(aSectionTable[KSectionData]);
	}

	if (aSectionTable[KSectionImport].iLoadedVma)
	{
		DPRINT((HERE("CGZipPEFileReader::LoadSectionsL - import section")));
		GunzipSectionL(aSectionTable[KSectionImport]);
	}

	if (aSectionTable[KSectionReloc].iLoadedVma)
	{
		DPRINT((HERE("CGZipPEFileReader::LoadSectionsL - reloc section")));
		GunzipSectionL(aSectionTable[KSectionReloc]);
	}

	iGzipReader.CheckEnd();

	DPRINT((HERE("CGZipPEFileReader::LoadSectionsL - done")));
}

#else



/*******************************************
 *
 * PE FileReader
 *
 * This supports PE files as output by 
 * the GNU linker. For testing purposes only.
 *
 *******************************************/

#define PE_INT32(var) (*((int*) (&(var)[0])))
#define PE_INT16(var) (*((short*) (&(var)[0])))
#define PE_UINT32(var) (*((unsigned*) (&(var)[0])))
#define PE_UINT16(var) (*((unsigned short*) (&(var)[0])))


CFileReaderBase* CreateFileReaderL()
{
	return CPEFileReader::NewL();
}


CPEFileReader* CPEFileReader::NewL()
{
	CPEFileReader*	reader = new(ELeave)CPEFileReader;
	CleanupStack::PushL(reader);
	reader->ConstructL();
	CleanupStack::Pop(reader);
	return reader;
}

CPEFileReader::~CPEFileReader()
{
	DPRINT((HERE("CPEFileReader::~CPEFileReader")));

	iFile.Close();
	iFs.Close();

	DPRINT((HERE("CPEFileReader::~CPEFileReader - done")));
}


void CPEFileReader::ConstructL()
{
	DPRINT((HERE("CPEFileReader::ConstructL")));

	PanicIfError(iFs.Connect(), ECompXLPanicCouldNotConnectToFileServer);

	TFileName	me = RProcess().FileName();
	RFile	file;
	TInt err = file.Open(iFs, me, 0);
	if (err)
	{
		DPRINT((HERE("ERROR opening me: %S, err=%d"), &me, err));
	}
	else
	{
		DPRINT((HERE("Successfully opened me: %S"), &me));
	}

	
#ifdef __WINS__
	PanicIfError(iFile.Open(iFs, _L("c:\\system\\programs\\hello.bin"), 0), ECompXLPanicCouldNotOpenOwnImageFile);
#else
	PanicIfError(iFile.Open(iFs, _L("d:\\system\\programs\\hello.bin"), 0), ECompXLPanicCouldNotOpenOwnImageFile);
#endif

	DPRINT((HERE("CPEFileReader::ConstructL - done")));
}

void CPEFileReader::LoadHeadersL(TImageInfo& aImageInfo, TSectionTableEntry* aSectionTable)
{
	DPRINT((HERE("CPEFileReader::ReadHeadersL")));

	TPtr8	p1((TUint8*)&iExtPeHeader, sizeof(iExtPeHeader));
	PanicIfError(iFile.Read(p1), ECompXLPanicCouldNotReadExtPEHeader);

	TPtr8	p2((TUint8*)&iPeOutHeader, sizeof(iPeOutHeader));
	PanicIfError(iFile.Read(p2), ECompXLPanicCouldNotReadAoutPEHeader);

	iNoOfSections = (TInt) PE_UINT16(iExtPeHeader.f_nscns);
	TPtr8	p3((TUint8*)&iSectionHeaders[0], sizeof(external_scnhdr) * iNoOfSections);
	PanicIfError(iFile.Read(p3), ECompXLPanicCouldNotReadPESectionHeaders);

	aImageInfo.iImageBase = (TUint8*) PE_UINT32(iPeOutHeader.ImageBase);
	aImageInfo.iImageEntyPoint = aImageInfo.iImageBase + PE_UINT32(iPeOutHeader.standard.entry);

	BuildSectionTableL(aSectionTable);


	DPRINT_DUMP_HEADERS();

	DPRINT((HERE("CPEFileReader::ReadHeadersL - done")));
}


void CPEFileReader::LoadSectionsL(const TSectionTableEntry* aSectionTable)
{
	DPRINT((HERE("CPEFileReader::LoadSectionsL")));

	DPRINT((HERE("CPEFileReader::LoadSectionsL - loading text @ 0x%x"), aSectionTable[KSectionCode].iLoadedVma));
	if (aSectionTable[KSectionCode].iLoadedVma)
	{
		external_scnhdr*		section = Section(KSectionCode);
		PanicIfNull(section, ECompXLPanicNoPECodeSection);
		TInt	size = PE_INT32(iPeOutHeader.standard.tsize);
		PanicIfError(iFile.Seek(ESeekStart, PE_INT32(section->s_scnptr)), ECompXLPanicCouldNotSeekToPECodeSection);
		TPtr8	p(aSectionTable[KSectionCode].iLoadedVma, size);
		PanicIfError(iFile.Read(p), ECompXLPanicCouldNotReadPECodeSection);
	}

	if (aSectionTable[KSectionData].iLoadedVma)
	{
		external_scnhdr*		section = Section(KSectionData);
		PanicIfNull(section, ECompXLPanicNoPEDataSection);
		TInt	size = PE_INT32(section->s_size);
		DPRINT((HERE("CPEFileReader::LoadSectionsL - loading data @ 0x%x"), aSectionTable[KSectionData].iLoadedVma));
		PanicIfError(iFile.Seek(ESeekStart, PE_INT32(section->s_scnptr)), ECompXLPanicCouldNotSeekToPEDataSection);
		TPtr8	p(aSectionTable[KSectionData].iLoadedVma, size);
		PanicIfError(iFile.Read(p), ECompXLPanicCouldNotReadPEDataSection);
	}

	if (aSectionTable[KSectionImport].iLoadedVma)
	{
		external_scnhdr*		section = Section(KSectionImport);
		PanicIfNull(section, ECompXLPanicNoPEImportSection);
		TInt	size = PE_INT32(section->s_size);
		DPRINT((HERE("CPEFileReader::LoadSectionsL - loading import @ 0x%x"), aSectionTable[KSectionImport].iLoadedVma));
		PanicIfError(iFile.Seek(ESeekStart, PE_INT32(section->s_scnptr)), ECompXLPanicCouldNotSeekToPEImportSection);
		TPtr8	p(aSectionTable[KSectionImport].iLoadedVma, size);
		PanicIfError(iFile.Read(p), ECompXLPanicCouldNotReadPEImportSection);
	}

	if (aSectionTable[KSectionReloc].iLoadedVma)
	{
		external_scnhdr*		section = Section(KSectionReloc);
		PanicIfNull(section, ECompXLPanicNoPERelocSection);
		TInt	size = PE_INT32(section->s_size);
		DPRINT((HERE("CPEFileReader::LoadSectionsL - loading reloc @ 0x%x"), aSectionTable[KSectionReloc].iLoadedVma));
		PanicIfError(iFile.Seek(ESeekStart, PE_INT32(section->s_scnptr)), ECompXLPanicCouldNotSeekToPERelocSection);
		TPtr8	p(aSectionTable[KSectionReloc].iLoadedVma, size);
		PanicIfError(iFile.Read(p), ECompXLPanicCouldNotReadPERelocSection);
	}

	DPRINT((HERE("CPEFileReader::LoadSectionsL - done")));
}



external_scnhdr* CPEFileReader::Section(TSectionIndex aSectionIndex)
{
	TInt sectionIndex = iSectionHeadersIndex[aSectionIndex];
	
	return (sectionIndex>=0 && sectionIndex<iNoOfSections) ?
		&iSectionHeaders[sectionIndex] : NULL;
}


void CPEFileReader::BuildSectionTableL(TSectionTableEntry* aSectionTable)
{
	DPRINT((HERE("CPEFileReader::BuildSectionTableL")));

	for (TInt j=0 ; j<KMaxSectionHeaders ; j++)
			iSectionHeadersIndex[j] = -1;

	for (TInt i=0 ; i<iNoOfSections ; i++)
	{
		TPtrC8 sectionName((const TUint8*) &iSectionHeaders[i].s_name[0]);
		if (sectionName.Compare(_L8(".text")) == 0)
		{
			BuildSectionTableEntryL(iSectionHeaders[i], aSectionTable[KSectionCode]);
			iSectionHeadersIndex[KSectionCode] = iCodeSectionSize;
		}
		else if (sectionName.Compare(_L8(".data")) == 0)
		{
			BuildSectionTableEntryL(iSectionHeaders[i], aSectionTable[KSectionData]);
			iSectionHeadersIndex[KSectionData] = i;
		}
		else if (sectionName.Compare(_L8(".bss")) == 0)
		{
			BuildSectionTableEntryL(iSectionHeaders[i], aSectionTable[KSectionBss]);
			iSectionHeadersIndex[KSectionBss] = i;
		}
		else if (sectionName.Compare(_L8(".idata")) == 0)
		{
			BuildSectionTableEntryL(iSectionHeaders[i], aSectionTable[KSectionImport]);
			iSectionHeadersIndex[KSectionImport] = i;
		}
		else if (sectionName.Compare(_L8(".reloc")) == 0)
		{
			BuildSectionTableEntryL(iSectionHeaders[i], aSectionTable[KSectionReloc]);
			iSectionHeadersIndex[KSectionReloc] = i;
		}
	}

	DPRINT((HERE("CPEFileReader::BuildSectionTableL - done")));
}

void CPEFileReader::BuildSectionTableEntryL(const external_scnhdr& aExternalSection, TSectionTableEntry& aSectionTableEntry)
{
	TUint8*		imageBase = (TUint8*) PE_UINT32(iPeOutHeader.ImageBase);
	TUint		offset = PE_UINT32(aExternalSection.s_vaddr);
	aSectionTableEntry.iSize = PE_INT32(aExternalSection.s_size);
	aSectionTableEntry.iImageVma = imageBase + offset;
	aSectionTableEntry.iStartBlock = offset / KPageSize;
	aSectionTableEntry.iLastBlock = (aSectionTableEntry.iStartBlock + aSectionTableEntry.iSize + KPageSize - 1) / KPageSize;
}



#ifdef WANT_DPRINT
void CPEFileReader::DumpHeaders()
{
	RDebug::Print(_L("ImageBase= 0x%x"), PE_UINT32(iPeOutHeader.ImageBase));
	RDebug::Print(_L("ImageSize= 0x%x"), PE_UINT32(iPeOutHeader.SizeOfImage));
	RDebug::Print(_L("CodeSize= 0x%x"), PE_UINT32(iPeOutHeader.standard.tsize));
	RDebug::Print(_L("InitialisedDataSize= 0x%x"), PE_UINT32(iPeOutHeader.standard.dsize));
	RDebug::Print(_L("UninitialisedDataSize= 0x%x"), PE_UINT32(iPeOutHeader.standard.bsize));
	for (TInt i=0 ; i<iNoOfSections ; i++)
	{
		TBuf<32>	sectionName;
		TPtrC8	p((TUint8*) iSectionHeaders[i].s_name);
		sectionName.Copy(p);
		RDebug::Print(_L("Section: %S"), &sectionName);
		RDebug::Print(_L("  phys addres  : 0x%08x"), PE_UINT32(iSectionHeaders[i].s_paddr));
		RDebug::Print(_L("  virt addres  : 0x%08x"), PE_UINT32(iSectionHeaders[i].s_vaddr));
		RDebug::Print(_L("  size         : 0x%08x"), PE_UINT32(iSectionHeaders[i].s_size));
		RDebug::Print(_L("  section file : 0x%08x"), PE_UINT32(iSectionHeaders[i].s_scnptr));
		RDebug::Print(_L("  reloc file   : 0x%08x"), PE_UINT32(iSectionHeaders[i].s_relptr));
		RDebug::Print(_L("  num relocs   : %d"), PE_UINT16(iSectionHeaders[i].s_nreloc));
	}
}
#endif

#endif