/*******************************************************************
 *
 *	File:		CImageLoader.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 2002, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include <e32def.h>
#include <e32uid.h>
#include <f32file.h>
#include "debug.h"

#include "CImageLoader.h"
#include "Panic.h"

#define ILT_LINK_BY_ORDINAL (0x80000000u)

#define ROUND_DWORD(x) (((x) + 3 ) & ~(0x3))

/*
 * Relocation types as defined by the MSDN SDK, section
 * "Microsoft Portable Executable and Common Object File Format Specification"
 *
 * It seems that only 32-bit relocation types are used.
 */
#define IMAGE_REL_BASED_ABSOLUTE	0		// used for padding, i.e., ignore
#define IMAGE_REL_BASED_HIGHLOW		3		// relocation in 32 its address


void Panic(TInt aPanic)
{
	User::Panic(_L("ECompXL"), aPanic);
}



CImageLoader* CImageLoader::NewLC()
{
	CImageLoader*	loader = new(ELeave)CImageLoader;
	CleanupStack::PushL(loader);
	loader->ConstructL();
	return loader;
}

CImageLoader::~CImageLoader()
{
	DPRINT((HERE("CImageLoader::~CImageLoader")));

	delete iRelocBuf;
	iRelocBuf = NULL;

	delete iFileReader;
	iFileReader = NULL;

	DPRINT((HERE("CImageLoader::~CImageLoader - done")));
}


void CImageLoader::ConstructL()
{
	DPRINT((HERE("CImageLoader::ConstructL")));

	iFileReader = CreateFileReaderL();

	DPRINT((HERE("CImageLoader::ConstructL - done")));
}


TUint CImageLoader::LoadImageL()
{
	DPRINT((HERE("CImageLoader::LoadImageL")));

	iFileReader->LoadHeadersL(iImageInfo, iSectionTable);
	DUMP_HEADERS();
	CreateChunksL();
	iFileReader->LoadSectionsL(iSectionTable);
	ImportLibrariesL();
	RelocateL();

	TUint	entryPointloadedVma = ImageVmaToLoadedVma((TUint32)iImageInfo.iImageEntyPoint);

	DPRINT((HERE("CImageLoader::LoadImageL - done, entry=0x%08x"), entryPointloadedVma));

	return entryPointloadedVma;
}


void CImageLoader::CreateChunksL()
{
	TInt	codeSize = iSectionTable[KSectionCode].iSize;
	PanicIfError(iCodeChunk.CreateLocalCode(codeSize, codeSize), ECompXLPanicCreateCodeChunk);
	iSectionTable[KSectionCode].iLoadedVma = iCodeChunk.Base();

	TInt	dataSize = ROUND_DWORD(iSectionTable[KSectionData].iSize) + ROUND_DWORD(iSectionTable[KSectionBss].iSize) + ROUND_DWORD(iSectionTable[KSectionImport].iSize);
	PanicIfError(iDataChunk.CreateLocal(dataSize, dataSize), ECompXLPanicCreateCodeChunk);
	TUint8*		base = iDataChunk.Base();
	if (iSectionTable[KSectionData].iSize)
	{
		iSectionTable[KSectionData].iLoadedVma = base;
		base += ROUND_DWORD(iSectionTable[KSectionData].iSize);
	}
	if (iSectionTable[KSectionBss].iSize)
	{
		iSectionTable[KSectionBss].iLoadedVma = base;
		Mem::FillZ(base, iSectionTable[KSectionBss].iSize);
		base += ROUND_DWORD(iSectionTable[KSectionBss].iSize);
	}
	if (iSectionTable[KSectionImport].iSize)
	{
		iSectionTable[KSectionImport].iLoadedVma = base;
	}

	iRelocBuf = HBufC8::NewL(ROUND_DWORD(iSectionTable[KSectionReloc].iSize));
	iSectionTable[KSectionReloc].iLoadedVma = (TUint8*) iRelocBuf->Ptr();
}


static TUint AtoI(const TDesC& aString)
{
	TInt	length = aString.Length();
	TUint	value = 0;
	for (TInt i=0 ; i<length ; i++)
	{
		value <<= 4;
		TUint16	c = aString[i];
		if ((c >= '0') && (c <='9'))
			value |= c - '0';
		else if ((c >= 'a') && (c <='f'))
			value |= c - 'a' + 10;
		else if ((c >= 'A') && (c <='F'))
			value |= c - 'A' + 10;
	}
	return value;
}

void CImageLoader::GetPureDllNameAndUid(const TDesC& aMangledDllName, TDes& aDllName, TUid& aUid)
{
	DPRINT((HERE("CImageLoader::GetPureDllNameAndUid, %S"), &aMangledDllName));

	TInt	posLeft = aMangledDllName.Locate('[');
	TInt	posRight = aMangledDllName.Locate(']');
	if ((posLeft > 0) && (posRight > posLeft))
	{
		aDllName.Copy(aMangledDllName.Left(posLeft));
		aDllName.Append(aMangledDllName.Right(aMangledDllName.Length() - posRight - 1));

		aUid.iUid = AtoI(aMangledDllName.Mid(posLeft + 1, posRight-posLeft -1));
	}
	else
	{
		// hmm, normally libraries are in the format: "somelib.dll[0x12345678]"
		// it seems though that sometimes the UID is omitted and only the dll name is given
		aDllName = aMangledDllName;
		aUid = KNullUid; 
	}

	DPRINT((HERE("CImageLoader::GetPureDllNameAndUid, - done dll=%S, uid=0x%x"), &aDllName, aUid.iUid));
}

TFileName CImageLoader::DirectoryImRunningFrom()
{
#ifdef ECOMPXL_EXE
	#ifdef __WINS__
		TFileName	me = _L("c:\\system\\programs\\hello.exe");
	#else
		TFileName	me = RProcess().FileName();
	#endif
#else
	#ifdef __WINS__
		TFileName	me = _L("c:\\system\\apps\\hello\\hello.app");
	#else
		TFileName	me;
		Dll::FileName(me);
	#endif
#endif
	TParsePtrC	parse(me);
	return parse.DriveAndPath();
}


void CImageLoader::ImportLibraryL(
		const TDesC&					aDllName, 
		const TUidType&					aUidType, 
		const import_dir_table_entry*	aIdte
		)
{
	DPRINT((HERE("CImageLoader::ImportLibraryL, %S"), &aDllName));

		/*
		 * Load library and fixup the Import Address Table.
		 * Note that we don't close the library. This is intentional as this process needs
		 * a reference to it for as long as it running. The kernel will cleanup for us when 
		 * the process is gone.
		 */
	RLibrary	lib;
	TInt err = lib.Load(aDllName, aUidType);
	if (err != KErrNone)
	{
		TParsePtrC		dllName(aDllName);
		TFindFile		findDll(iFileReader->iFs);
		err = findDll.FindByDir(dllName.NameAndExt(), _L("\\system\\libs\\"));
		if (err == KErrNone)
		{
			DPRINT((HERE("  found it in system libs")));
			err = lib.Load(findDll.File());
		}
		else
		{
			TFileName	myDir = DirectoryImRunningFrom();
			err = findDll.FindByDir(dllName.NameAndExt(), myDir);
			if (err == KErrNone)
			{
				RDebug::Print(_L("  found it in my app or exe dir"));
				err = lib.Load(findDll.File());
			}
		}
	}
	else
	{
		DPRINT((HERE("  system found the dll for us")));
	}
	PanicIfError(err, ECompXLPanicLoadLibrary);
	
	TUint32*	importAdrressTable = (TUint32*) (iSectionTable[KSectionImport].
	iLoadedVma + (aIdte->idte_first_thunk - (TUint32) (iSectionTable[KSectionImport].iImageVma - iImageInfo.iImageBase)));
	DPRINT((HERE("CImageLoader::ImportLibraryL, Import Address Table=0x%x"), importAdrressTable));

	TInt	count = 0;
	while(*importAdrressTable)
	{
		DPRINT((HERE("CImageLoader::ImportLibraryL, Import Address=0x%x"), *importAdrressTable));
		if (!(*importAdrressTable & ILT_LINK_BY_ORDINAL))
			Panic(ECompXLPanicNotLinkByOrdinal);
		TInt ordinal = *importAdrressTable & ~ILT_LINK_BY_ORDINAL;

		*importAdrressTable++ = (TUint32) lib.Lookup(ordinal);
		count++;
	}
	DPRINT((HERE("CImageLoader::ImportLibraryL, - done import count=%d"), count));
}


void CImageLoader::ImportLibrariesL()
{
	DPRINT((HERE("CImageLoader::ImportLibrariesL")));

	TUint8*		sectionData = iSectionTable[KSectionImport].iLoadedVma;
	import_dir_table_entry*		idte = (import_dir_table_entry*) sectionData;
	do 
	{
		TBuf<64>	mangledDllName;
		TBuf<64>	dllName;
		TUid		dllUid;

		TUint8*		c_dllName = sectionData + (idte->idte_dll_name - (TUint32) (iSectionTable[KSectionImport].iImageVma - iImageInfo.iImageBase));
		TPtrC8		p(c_dllName);
		mangledDllName.Copy(p);
		GetPureDllNameAndUid(mangledDllName, dllName, dllUid);
		
		TUidType	uidType(KDynamicLibraryUid, KSharedLibraryUid, dllUid);
		ImportLibraryL(dllName, uidType, idte);

		idte++;

	} while(idte->idte_dll_name != 0);

	DPRINT((HERE("CImageLoader::ImportLibrariesL - done")));
}



extern "C" {
TInt _E32Dll(TDllReason);
}

void CImageLoader::RelocateL()
{
	DPRINT((HERE("CImageLoader::RelocateL")));

	TInt				size = iSectionTable[KSectionReloc].iSize;
	fixup_block*		fixupBlock = (fixup_block*) iSectionTable[KSectionReloc].iLoadedVma;
	TInt				count = 0;
	TInt				section = 0;
	TInt				blocks=0;
	TSectionTableEntry*	relocTableEntry = &iSectionTable[0];

#ifdef ECOMPXL_APP
	TUint				originalE32Dll = (TUint) iSectionTable[KSectionCode].iImageVma;
#endif

	while (count < size)
	{
		TInt		block = fixupBlock->page / KPageSize;
		if ((block == 0) || (fixupBlock->block_size == 0))
		{
			DPRINT((HERE("NULL block, end-of-table")));
			break;
		}
		while (!((block>= relocTableEntry->iStartBlock) && (block<relocTableEntry->iLastBlock) && (section<KMaxSections)))
		{
			relocTableEntry++;
			section++;
			DPRINT((HERE("After %d blocks, moving to section %d"), blocks, section));
		}
		if (section > KMaxSections)
			Panic(ECompXLPanicRelocOutOfRange);

		TUint16*	fixup = (TUint16*) (fixupBlock+1);
		TInt		noOfFixups = (fixupBlock->block_size - sizeof(fixup_block)) / sizeof(TUint16);
		DPRINT((HERE("Section %d, block=%8d, fixups=%d, blksize=%d"), section, block, noOfFixups, fixupBlock->block_size));
		while (noOfFixups--)
		{
			TInt	type = 	(*fixup & 0xf000) >> 12;
			TInt	offset = *fixup & 0x0fff;

			if (type == IMAGE_REL_BASED_HIGHLOW)
			{
#ifdef WANT_DPRINT
				TUint32*		relocImageVma = (TUint32*) (fixupBlock->page + offset + iImageInfo.iImageBase);
#endif
				TUint32*		relocLoadedVma = (TUint32*)  (((block - relocTableEntry->iStartBlock) << 12) + relocTableEntry->iLoadedVma + offset);
				TUint32			relocRefersToImageVma  = *relocLoadedVma;

				TUint32			relocRefersToLoadedVma = ImageVmaToLoadedVma(relocRefersToImageVma);
				DPRINT((HERE("Reloc: [0x%08x]=0x%08x --> [0x%08x]=0x%08x"),relocImageVma, relocRefersToImageVma, relocLoadedVma, relocRefersToLoadedVma));


#ifdef ECOMPXL_APP
		/*
		 * The relocated address of E32Dll is used by Symbian OS kernel as a unique key for the 
		 * loaded dll.However, the kernel now knows this dll by the ECompXL's loader E32Dll address. 
		 * We therefor need to ensure that all references to the .dll's E32Dll are fixed-up to 
		 * ECompXL's loader E32Dll address.
		 */
if ((relocRefersToImageVma == originalE32Dll) || (relocRefersToImageVma == (originalE32Dll+1)))
   *relocLoadedVma = ((TInt)_E32Dll);
else        // no normal fix-up

#endif
				*relocLoadedVma = relocRefersToLoadedVma;
			}
			else if (type != IMAGE_REL_BASED_ABSOLUTE)
			{
				DPRINT((HERE("Unsupported relocation type=%d"), type));
				Panic(ECompXLPanicUnsupportedRelocType);
			}
			fixup++;
		}
		count += fixupBlock->block_size;
		fixupBlock = (fixup_block*) (((TUint8*)fixupBlock) + fixupBlock->block_size);
		blocks++;
	}

	delete iRelocBuf;
	iRelocBuf = NULL;

		// make sure we flush the icache & dcache
	User::IMB_Range(iCodeChunk.Base(), iCodeChunk.Base()+iCodeChunk.Size());

	DPRINT((HERE("CImageLoader::RelocateL - done, fixup blocks=%d"), blocks));
}


#ifdef WANT_DPRINT
void CImageLoader::DumpSection(const TSectionTableEntry& aSection)
{
	RDebug::Print(_L("  size       : 0x%08x"), aSection.iSize);
	RDebug::Print(_L("  ImageVma   : 0x%08x"), aSection.iImageVma);
	RDebug::Print(_L("  startblock : 0x%d"), aSection.iStartBlock);
	RDebug::Print(_L("  lastblock  : 0x%d"), aSection.iLastBlock);
}

void CImageLoader::DumpHeaders()
{
	RDebug::Print(_L("ImageBase   = 0x%08x"), iImageInfo.iImageBase);
	RDebug::Print(_L("EntryPointe = 0x%08x"), iImageInfo.iImageEntyPoint);
	
	RDebug::Print(_L("Section: code"));
	DumpSection(this->iSectionTable[KSectionCode]);
	RDebug::Print(_L("Section: data"));
	DumpSection(this->iSectionTable[KSectionData]);
	RDebug::Print(_L("Section: bss"));
	DumpSection(this->iSectionTable[KSectionBss]);
	RDebug::Print(_L("Section: import"));
	DumpSection(this->iSectionTable[KSectionImport]);
	RDebug::Print(_L("Section: reloc"));
	DumpSection(this->iSectionTable[KSectionReloc]);
}
#endif
