/*******************************************************************
 *
 *	File:		ECompXLTran.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 2002, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <process.h>
#include <io.h>
#include <time.h>

#include "ECompXLWin32Common.h"

#define MAX_ARGC		64
#define MAX_CMD			1024
#define EXE_UID			0x1000007a
#define PATCH_DATA_SIZE	(64 * 1024)

//#define WANT_DPRINT 1
#ifdef WANT_DPRINT
#define DPRINT(x) printf x
#else
#define DPRINT(x)
#endif


static const char* FileNameOnly(const char* aFileName)
{
	char*	pc = strrchr(aFileName, '\\');
	if (pc)
		return pc+1;
	return aFileName;
}

static unsigned GetUid(int aArgc, char* aArgv[], int aUid)
{
	char	uidStr[32];
	sprintf(uidStr, "-uid%d", aUid);
	int		uidStrLen = strlen(uidStr);

	unsigned	uid = 0;
	for (int i=1 ; i<aArgc ; i++)
	{
		if (strnicmp(aArgv[i], uidStr, uidStrLen) == 0)
		{
			sscanf(aArgv[i] + uidStrLen + 1, "0x%x", &uid);
			break;
		}
	}
	return uid;
}

static void GetBinGzImageName(const char* aImageName, char* aBinGzImage)
{
	strcpy(aBinGzImage, aImageName);
	char*	pc = strrchr(aBinGzImage, '.');
	if (pc)
		strcpy(pc, ".bin.gz");
	else
		strcat(pc, ".bin.gz");
}


static char copyBuf[128*1024];
static void CopyFile(const char* aSrc, const char* aDst, bool aAppend, void* aExtraData, int aExtraDataSize)
{
	FILE*	fin = fopen(aSrc, "rb");
	FILE*	fout = fopen(aDst, aAppend ? "ab+" : "wb");
	if (!fin || !fout)
	{
		fprintf(stderr, "\nECompXL: could not copy loader image %s to %s\nBailing out\n\n", aSrc, aDst);
		exit(-1);
	}
	while (!feof(fin))
	{
		unsigned readCount = fread(copyBuf, 1, sizeof(copyBuf), fin);
		if (fwrite(copyBuf, 1, readCount, fout) != readCount)
		{
			fprintf(stderr, "\nECompXL: could not copy loader image %s to %s\nBailing out\n\n", aSrc, aDst);
			exit(-1);
		}
	}
	if (aExtraData)
	{
		if (fwrite(aExtraData, 1, aExtraDataSize, fout) != (unsigned) aExtraDataSize)
		{
			fprintf(stderr, "\nECompXL: could not copy extra data to image %s\nBailing out\n\n", aDst);
			exit(-1);
		}
	}

	fclose(fin);
	fclose(fout);
}


static void GetLoaderImage(char* aLoaderImage, bool aIsExe)
{
	char* path = strdup(getenv("PATH"));
	
	char*	pc = path;
	char*	next;
	do
	{
		next = strchr(pc, ';');
		if (next)
			*next = 0;

		sprintf(aLoaderImage, "%s\\%s", pc, aIsExe ? KECompXLExe : KECompXLApp);
//		DPRINT(("Looking for loader: %s\n",aLoaderImage));
		if (access(aLoaderImage, 0) == 0)
		{
			free(path);
			return;
		}
		aLoaderImage[0] = 0;
		pc = next + 1;
	} while (next);

	fprintf(stderr, "\nECompXL: could not find loader image (%s) in PATH\n\n", aIsExe ? KECompXLExe : KECompXLApp);
	exit(-1);
}



void RunPetran(int aArgc, char* aArgv[])
{
	char*	petranArgv[MAX_ARGC];
	petranArgv[0] = (char*) KOriginalPetranExe;

		/*
		 * Parameters for petran are the same as the orignal, with the following 2 excpetions:
		 * 1) ommit first parameter (we only want petran to update the e32 image with the
		 * right uids, stack size etc.
		 * 2) If the stack size indicates 0x80000000 then ommit the parameter to get
		 * the default stack size, otherwise pass stack parameter with high bit set.
		 */
	int argc = 1;
	for (int i=2 ; i<aArgc ; i++)
	{
		if (strnicmp(aArgv[i], "-stack", 6) == 0)
		{
			unsigned	stackSize = 0;
			sscanf(aArgv[++i], "0x%x", &stackSize);
			if (stackSize != 0x80000000)
			{
				stackSize &= ~0x80000000;
				char*	stackStr = (char*) malloc(64);
				sprintf(stackStr, "0x%x", stackSize);
				petranArgv[argc++] = aArgv[i-1];
				petranArgv[argc++] = stackStr;
			}
		}
		else
			petranArgv[argc++] = aArgv[i];
	}
	petranArgv[argc] = NULL;
	if (_spawnvp(_P_WAIT, KOriginalPetranExe, petranArgv) != 0)
	{
		fprintf(stderr, "\nECompXL: could not run EPOC's Petran, bailing out\n");
		fprintf(stderr, "Please consult the readme on how to integrate ECompXL with the EPOC toolchain\n\n");
		exit(-1);
	}
}


static int GetFileSize(const char* aFileName)
{
	struct _stat	statFile;
	if (_stat(aFileName, &statFile) != 0)
	{
		fprintf(stderr, "ERROR: Could not file size for %s\n", aFileName);
		exit(-1);
	}
	return statFile.st_size;
}

static int GetCompressionRatio(const char* aSrc, const char* aDst)
{
	float srcSize = (float) GetFileSize(aSrc);
	float dstSize = (float) GetFileSize(aDst);

	float ratio = 100 - 100*(dstSize / srcSize);
	return (int) ratio;
}

static void RunPE2ECompXL(const char* aPEFile)
{
	char	cmd[MAX_CMD];

	sprintf(cmd, "%s %s", KPE2ECompXL, aPEFile);
	DPRINT(("Executing %s\n",cmd));
	if ( _spawnlp(_P_WAIT , KPE2ECompXL, KPE2ECompXL, aPEFile, "/q", NULL) != 0)
	{
		fprintf(stderr, "\nECompXL: %s failed, bailing out\n\n", KPE2ECompXL);
		exit(-1);
	}
}


static void PatchLoaderImageData(char* aData, int aSize)
{
	char*	pattern = (char*) "POMPOM_AND_PATCHY_ARE_MY_CATS";
	int		patternLength = strlen(pattern);
	
	int count = aSize - patternLength;

	while (count--)
	{
		if (memcmp(aData, pattern, patternLength) == 0)
			break;
		aData++;
	}
	if (count < 0)
	{
		fprintf(stderr, "\nECompXL: bailing out, no patch pattern in loader\n\n");
		exit(-1);
	}

	time_t	timeNow;
	time(&timeNow);
	sprintf(aData, "%ld", timeNow);
}


void NewCheckSum(char * aData);

static void PatchLoderImage(const char* aLoaderImage)
{
	FILE*	file = fopen(aLoaderImage, "r+b");
	if (!file)
	{
		fprintf(stderr, "\nECompXL: could not open loader for patching\n  loader=%s\n\n", aLoaderImage);
		exit(-1);
	}
	char* data = (char*)malloc(PATCH_DATA_SIZE);
	if (!data)
	{
		fprintf(stderr, "\nECompXL: not enough memory to patch loader image\n  loader=%s\n\n", aLoaderImage);
		exit(-1);
	}

	int		done = 0;
	int		size = 0;
	int		maxsize = PATCH_DATA_SIZE;
	char*	d = data;
	do
	{
		int s = fread(d, 1, maxsize, file);
		
		maxsize -= s;
		d += s;
		size += s;

		done = ((s==0) || (maxsize == 0));
	} while (!done);

	if (size <=0)
	{
		fprintf(stderr, "\nECompXL: could not read loader for patching\n  loader=%s\n\n", aLoaderImage);
		exit(-1);
	}

	PatchLoaderImageData(data, size);
	NewCheckSum(data);

	fseek(file, 0, SEEK_SET);

	done = 0;
	d = data;
	do
	{
		int s = fwrite(d, 1, size, file);
		
		d += s;
		size -= s;
	
		done = ((s==0) || (size==0));
	} while (!done);

	if (size>0)
	{
		fprintf(stderr, "\nECompXL: could not write loader for patching\n  loader=%s\n\n", aLoaderImage);
		exit(-1);
	}

	free(data);
	fclose(file);
}


static void GenerateCompressedImage(int aArgc, char* aArgv[])
{
	char	tmpOutputName[_MAX_PATH];
	char	loaderImage[_MAX_PATH];
	char	binGzImage[_MAX_PATH];

	unlink(aArgv[2]);
	char* origOutputName = aArgv[2];
	sprintf(tmpOutputName, "%s.tmp", aArgv[2]);
	aArgv[2] = tmpOutputName;

	bool isExe = GetUid(aArgc, aArgv, 1) == EXE_UID;
	GetLoaderImage(loaderImage, isExe);
	GetBinGzImageName(aArgv[1], binGzImage);

	printf("  Running %s to create %s\n", KPE2ECompXL, FileNameOnly(binGzImage));
	RunPE2ECompXL(aArgv[1]);

		// copy loaderImage to destination and run petran on it
	printf("  Running EPOC Petran on loader image %s\n", FileNameOnly(loaderImage));
	CopyFile(loaderImage, aArgv[2], false, NULL, 0);
	PatchLoderImage(aArgv[2]);
	RunPetran(aArgc, aArgv);

	printf("ECompXL: glueing %s to loader image %s\n", FileNameOnly(binGzImage), FileNameOnly(loaderImage));
	int	loaderFileSize = GetFileSize(aArgv[2]);
	CopyFile(binGzImage, aArgv[2], true, &loaderFileSize, sizeof(loaderFileSize));
	if (rename(aArgv[2], origOutputName) != 0)
	{
		fprintf(stderr, "\nECompXL: could not remame %s to %s\n\n", aArgv[2], FileNameOnly(origOutputName));
		exit(-1);
	}
	int ratio = GetCompressionRatio(aArgv[1], origOutputName);
	printf("Successfully created compressed executable %s (compressed %d%%)\n\n", FileNameOnly(origOutputName), ratio);
}


int main(int aArgc, char* aArgv[])
{
#ifdef _DEBUG
	printf("***********************************\n");
	printf("*** DEBUG VERSION OF %s\n", aArgv[0]);
	printf("***********************************\n");
	_asm int 3;
#endif

	PrintBanner();
	if (aArgc > 1)
		GenerateCompressedImage(aArgc, aArgv);
	return 0;
}




