/*******************************************************************
 *
 *	File:		PE2ECompXL.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *	(c) Copyright 2002, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <process.h>
#include <io.h>

#define COFF_IMAGE_WITH_PE
#include "coff_pe/arm.h"
#include "coff_pe/pe.h"

#include "ECompXLWin32Common.h"
#include "ECompXLFileFormat.h"

#define WANT_DPRINT 1
#ifdef WANT_DPRINT
#define DPRINT(x) fprintf x
#else
#define DPRINT(x)
#endif

#define PE_INT32(var) (*((int*) (&(var)[0])))
#define PE_INT16(var) (*((short*) (&(var)[0])))
#define PE_UINT32(var) (*((unsigned*) (&(var)[0])))
#define PE_UINT16(var) (*((unsigned short*) (&(var)[0])))


typedef struct
{
	unsigned	edt_export_flags;
	unsigned	edt_time_stamp;
	unsigned	edt_version;
	unsigned	edt_name;			/* rva - relative to image base */
	unsigned	edt_base;			/* ordinal base */
	unsigned	edt_num_functions;	/* Number in the export address table */
	unsigned	edt_num_names;		/* Number in the name pointer table */
	unsigned	edt_ead_addr;		/* rva to the export address table */
	unsigned	edt_enp_addr;		/* rva to the Export Name Pointer Table */
	unsigned	edt_ot_addr;		/* rva to the Ordinal Table */
} export_dir_table_entry;



unsigned GetExportedOrdinalOne(FILE* aFile, const external_scnhdr& aSectionHeader, const char* aPEFile)
{
	unsigned	export = 0;
	int			sectionSize = PE_INT32(aSectionHeader.s_size);
	int			sectionFileOffset = PE_INT32(aSectionHeader.s_scnptr);
	char*	sectionData = (char*) malloc(sectionSize);
	if (sectionData)
	{
		if (fseek(aFile, sectionFileOffset, SEEK_SET) != 0)
		{
			fprintf(stderr, "ERROR: Could not seek to export section (%d) in file %s, error=%d\n", 
					sectionFileOffset, aPEFile, errno);
			exit (-1);
		}
		if (fread(sectionData, sectionSize, 1, aFile) != 1)
		{
			fprintf(stderr, "ERROR: Could not read export section in file %s, error=%d\n", 
					aPEFile, errno);
			exit (-1);
		}

		export_dir_table_entry*		edte = (export_dir_table_entry*) sectionData;

		unsigned*	rva = (unsigned*) (sectionData + (edte->edt_ead_addr - PE_INT32(aSectionHeader.s_vaddr)));
		export = rva[0];

		free (sectionData);
	}
	return export;
}

size_t my_fread( void *buffer, size_t size, size_t count, FILE *stream)
{
	size_t	orig_count = count;
	char*	buf = (char*)buffer;
	while (!feof(stream) && (count>0))
	{
		int blocks = fread(buf, size, count, stream);
		count -= blocks;
		buf += (blocks * size);
	}
	return orig_count - count;
}



char* GenerateOutFileName(const char* aPEFile)
{
	char*	outFileName = (char*) malloc(strlen(aPEFile) + 4);
	strcpy(outFileName, aPEFile);
	char*	pc = strrchr(outFileName, '.');
	if (pc)
		strcpy(pc, ".bin");
	else
		strcat(outFileName, ".bin");
	return outFileName;
}



static char	sBuf[128*1204];
static void CopySection(FILE* aInFile, const char* aPEFile, FILE* aOutFile, const char* aOutFileName, int aSectionFileOffset, int aSectionSize, const char* aSectionName)
{
	if (aSectionSize == 0)
		return;

	if (fseek(aInFile, aSectionFileOffset, SEEK_SET) != 0)
	{
		fprintf(stderr, "ERROR: Could not seek to %s section (%d) in file %s, error=%d\n", 
				aSectionName, aSectionFileOffset, aPEFile, errno);
		exit (-1);
	}
	int	remaining = aSectionSize;
	while (remaining > 0)
	{
		int	blockSize = (remaining < sizeof(sBuf)) ? remaining : sizeof(sBuf);
		if (fread(sBuf, blockSize, 1, aInFile) != 1)
		{
			fprintf(stderr, "ERROR: Could not read from %s section in file %s, error=%d\n", 
					aSectionName, aPEFile, errno);
			exit (errno);
		}
		if (fwrite(sBuf, blockSize, 1, aOutFile) != 1)
		{
			fprintf(stderr, "ERROR: Could not write %s section to file %s, error=%d\n", 
					aSectionName, aOutFileName, errno);
			exit (errno);
		}

		remaining -= blockSize;
	}
}

static void Compress(const char* aOutFileName)
{
	char	cmd[1024];
	sprintf(cmd, "%s -9 -f %s", KGZipExe, aOutFileName);
	if (_spawnlp(_P_WAIT, KGZipExe, KGZipExe, "-9", "-f", aOutFileName, NULL) != 0)
	{
		fprintf(stderr, "ERROR: Could not compress image %s\n", aOutFileName);
		exit(-1);
	}
}

static void PE2ECompXL(const char* aPEFile, int aQuiet)
{
	FILE*	inFile;
	FILE*	outFile;
	char*	outFileName = GenerateOutFileName(aPEFile);

	inFile = fopen(aPEFile, "rb");
	if (!inFile)
	{
		fprintf(stderr, "ERROR: Could not open PE image %s, error=%d\n", aPEFile, errno);
		exit (-1);
	}
	outFile = fopen(outFileName, "wb+");
	if (!outFile)
	{
		fprintf(stderr, "ERROR: Could not create compressed image file %s, error=%d\n", outFileName, errno);
		exit (errno);
	}


	PEAOUTHDR				peOutHeader;
	external_PE_filehdr		extPeHeader;
	external_scnhdr*		sectionHeaders;
	int						noOfSections;

	if (my_fread(&extPeHeader, sizeof(extPeHeader), 1, inFile) != 1)
	{
		fprintf(stderr, "ERROR: Could not read PE header from file %s, error=%d\n", aPEFile, errno);
		exit (errno);
	}
	if (my_fread(&peOutHeader, sizeof(peOutHeader), 1, inFile) != 1)
	{
		fprintf(stderr, "ERROR: Could not read AOUT header from file %s, error=%d\n", aPEFile, errno);
		exit (errno);
	}

	noOfSections = (int) PE_UINT16(extPeHeader.f_nscns);
	sectionHeaders = (external_scnhdr*) malloc(noOfSections * sizeof(external_scnhdr));
	if (my_fread(sectionHeaders, sizeof(external_scnhdr), noOfSections, inFile) != (unsigned) noOfSections)
	{
		fprintf(stderr, "ERROR: Could not read section headers from file %s, error=%d\n", aPEFile, errno);
		exit (errno);
	}

	TECompXLFileHeader	fileHeader;
	unsigned			codeSectionFileOffset;
	unsigned			dataSectionFileOffset;
	unsigned			importSectionFileOffset;
	unsigned			relocSectionFileOffset;

	memset(&fileHeader, 0, sizeof(fileHeader));
	fileHeader.iMagic = ECOMPXL_MAGIC;
	fileHeader.iVersion = KECompXLVersion1;
	fileHeader.iCodeSectionSize = PE_UINT32(peOutHeader.standard.tsize);
	fileHeader.iImageBaseVma = (unsigned) PE_UINT32(peOutHeader.ImageBase);
	fileHeader.iImageEntryPoint = fileHeader.iImageBaseVma + PE_UINT32(peOutHeader.standard.entry);

	for (int i=0 ; i<noOfSections ; i++)
	{
		if (strcmp(sectionHeaders[i].s_name, ".text") == 0)
		{
			fileHeader.iCodeSectionVma = fileHeader.iImageBaseVma + PE_UINT32(sectionHeaders[i].s_vaddr);
			codeSectionFileOffset = PE_INT32(sectionHeaders[i].s_scnptr);
		}
		else if (strcmp(sectionHeaders[i].s_name, ".bss") == 0)
		{
			fileHeader.iBssSectionSize = PE_UINT32(peOutHeader.standard.bsize);
			fileHeader.iBssSectionVma = fileHeader.iImageBaseVma + PE_UINT32(sectionHeaders[i].s_vaddr);
		}
		else if (strcmp(sectionHeaders[i].s_name, ".data") == 0)
		{
			fileHeader.iDataSectionSize = PE_UINT32(sectionHeaders[i].s_size);
			fileHeader.iDataSectionVma = fileHeader.iImageBaseVma + PE_UINT32(sectionHeaders[i].s_vaddr);
			dataSectionFileOffset = PE_INT32(sectionHeaders[i].s_scnptr);
		}
		else if (strcmp(sectionHeaders[i].s_name, ".reloc") == 0)
		{
			fileHeader.iRelocSectionSize = PE_UINT32(sectionHeaders[i].s_size);
			fileHeader.iRelocSectionVma = fileHeader.iImageBaseVma + PE_UINT32(sectionHeaders[i].s_vaddr);
			relocSectionFileOffset = PE_INT32(sectionHeaders[i].s_scnptr);
		}
		else if (strcmp(sectionHeaders[i].s_name, ".idata") == 0)
		{
			fileHeader.iImportSectionSize = PE_UINT32(sectionHeaders[i].s_size);
			fileHeader.iImportSectionVma = fileHeader.iImageBaseVma + PE_UINT32(sectionHeaders[i].s_vaddr);
			importSectionFileOffset = PE_INT32(sectionHeaders[i].s_scnptr);
		}
		else if (strcmp(sectionHeaders[i].s_name, ".edata") == 0)
		{
				/*
				 * We ignore the export section, apart from getting exported ordinal 1, which
				 * correspond to NewApplication for .apps
				 */
			unsigned exportedOrdinalOne = GetExportedOrdinalOne(inFile, sectionHeaders[i], aPEFile);
			fileHeader.iImageEntryPoint = fileHeader.iImageBaseVma + exportedOrdinalOne;
		}
		else
		{
			fprintf(stderr, "ERROR: unknown section %s in PE image, bailing out\n", sectionHeaders[i].s_name);
			exit (-1);
		}
	}
	if (fwrite(&fileHeader, sizeof(fileHeader), 1, outFile) != 1)
	{
		fprintf(stderr, "ERROR: Could not write file header to file %s, error=%d\n", outFileName, errno);
		exit (errno);
	}
	
	CopySection(inFile, aPEFile, outFile, outFileName, codeSectionFileOffset, fileHeader.iCodeSectionSize, "code");
	CopySection(inFile, aPEFile, outFile, outFileName, dataSectionFileOffset, fileHeader.iDataSectionSize, "data");
	CopySection(inFile, aPEFile, outFile, outFileName, importSectionFileOffset, fileHeader.iImportSectionSize, "import");
	CopySection(inFile, aPEFile, outFile, outFileName, relocSectionFileOffset, fileHeader.iRelocSectionSize, "reloc");

	fclose(inFile);
	fclose(outFile);

	Compress(outFileName);
	if (!aQuiet)
		printf("\n%s: sucessfully created %s.gz\n\n", KPE2ECompXL, outFileName);
}


static void Usage(const char* aExeName)
{
	printf("\nUsage: %s pe-filename [/q]\n", aExeName);
	printf("   Converts the PE executable into an ECompXL compressed image.\n");
	printf("   For example, d:\\aaa\\bbb\\foo.exe will be converted into d:\\aaa\\bbb\\foo.bin.gz\n");
	printf("   /q Surpress output messages\n");

	exit(-1);
}


int ParseParam(int aArgc, char* aArgv[], int& aQuiet)
{
	if (aArgc == 2)
		return 1;
	if (aArgc == 3)
	{
		for (int i=1 ; i<aArgc ; i++)
		{
			if (aArgv[i][0] == '/')
			{
				if (strnicmp(aArgv[i], "/q", 2) == 0)
					aQuiet = 1;
				else
					return 0;
			}
		}
		return 1;

	}
	return 0;
}


int main(int aArgc, char* aArgv[])
{
#ifdef _DEBUG
	printf("***********************************\n");
	printf("*** DEBUG VERSION OF %s\n", aArgv[0]);
	printf("***********************************\n");
	_asm int 3;
#endif

	int		quiet = 0;
	if (!ParseParam(aArgc, aArgv, quiet))
		Usage(aArgv[0]);

	PE2ECompXL(aArgv[1], quiet);
	return 0;
}

